﻿/*
 * Copyright 2011 hecke
 * 
 * This file is part of (G)VA18(B) multimeter protcol.
 * 
 * (G)VA18(B) multimeter protcol decoder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * (G)VA18(B) multimeter protcol decoder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with (G)VA18(B) multimeter protcol decoder. If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GVA18BMeterReader
{
    class GVA18BProtocolDecoder
    {
        private enum BitValues { /*convenience*/
                                    BIT0VAL = 1,
                                    BIT1VAL = 2,
                                    BIT2VAL = 4,
                                    BIT3VAL = 8,
                                    BIT4VAL = 16,
                                    BIT5VAL = 32,
                                    BIT6VAL = 64,
                                    BIT7VAL = 128
                               };

        /**
         * display value is encoded by maping the elements of the 7-segment
         * to the lower nibbles of byte groups [a,b] = {[1,2],[3,4],[5,6],[7,8]}:
         * 	     0a:00	
         * 0a:01       0b:00
	     *       0b:01	
         * 0a:02       0b:02
	     *       0b:03
         *       
         * by combining the nibbles to a byte we got a uniqe identifier 
         * for every symbol shown on the display.
         * 
         * for byte 1 bit 3 represents a leading minus (negative value)
         * for bytes 3,5,7 bit 3 indicates a leading decimal point
         * (means the encoded value is the first part of the fraction)
         */
        private enum DisplayValues:int
        {
            D_ZERO = 125,
            D_ONE = 5,
            D_TWO = 91,
            D_THREE = 31,
            D_FOUR = 39,
            D_FIVE = 62,
            D_SIX = 126,
            D_SEVEN = 21,
            D_EIGHT = 127,
            D_NINE = 63,
            D_EMPTY = 0,
            D_L = 104,        //for O.L -> overload
            D_MINUS = 2       //the - in this: -.-.-.- (not the leading minus in front of all digits)
        };

        /**
         * decoding of meter modes depends on values in bytes 0, 9, 10, 11, 12, 13
         * structure is: 
         * meter mode desc, unit, current type, byte0, byte9, byte10, byte11, byte12, byte13
         */
        private List<MeterModeDescriptor> meter_mode_descriptions = new List<MeterModeDescriptor> {         
            new MeterModeDescriptor("voltage","V", "AC", 0x09, 0x00, 0x00, 0x00, 0x04, 0x00),
            new MeterModeDescriptor("voltage","V", "DC", 0x05, 0x00, 0x00, 0x00, 0x04, 0x00),
            new MeterModeDescriptor("voltage","mV", "DC", 0x05, 0x00, 0x08, 0x00, 0x04, 0x00),
            new MeterModeDescriptor("resistance","\u2126", "", 0x01, 0x00, 0x00, 0x04, 0x00, 0x00),
            new MeterModeDescriptor("resistance","M\u2126", "", 0x01, 0x00, 0x02, 0x04, 0x00, 0x00),
            new MeterModeDescriptor("resistance", "K\u2126", "", 0x00, 0x00, 0x00, 0x00, 0x00, 0x00),
            new MeterModeDescriptor("diode test","V", "", 0x01, 0x01, 0x00, 0x00, 0x04, 0x00),
            new MeterModeDescriptor("continuity tester","", "", 0x01, 0x00, 0x01, 0x04, 0x00, 0x00),
            new MeterModeDescriptor("capacity","nF", "", 0x01, 0x04, 0x00, 0x08, 0x00, 0x00),
            new MeterModeDescriptor("capacity","uF", "", 0x01, 0x08, 0x00, 0x08, 0x00, 0x00),
            new MeterModeDescriptor("frequency","Hz", "", 0x01, 0x00, 0x00, 0x00, 0x02, 0x00),
            new MeterModeDescriptor("frequency","KHz", "", 0x01, 0x02, 0x00, 0x00, 0x02, 0x00),
            new MeterModeDescriptor("duty cycle","%", "", 0x01, 0x00, 0x04, 0x00, 0x00, 0x00),
            new MeterModeDescriptor("current","A", "AC", 0x09, 0x00, 0x00, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("current","A", "DC", 0x05, 0x00, 0x00, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("current","mA", "AC", 0x09, 0x00, 0x08, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("current","mA", "DC", 0x05, 0x00, 0x08, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("current","uA", "AC", 0x09, 0x08, 0x00, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("current","uA", "DC", 0x05, 0x08, 0x00, 0x00, 0x08, 0x00),
            new MeterModeDescriptor("temperature","°C", "", 0x01, 0x00, 0x00, 0x00, 0x00, 0x02)            
        };

        private int state = 0;  /*number of last seen byte in the stream*/
        private byte[] data = new byte[14];
        private bool valid_data = false;
          
        private bool auto_range;
        private bool max_value;
        private bool min_value;
        private bool data_hold;
        private bool relative_measurement;
        
        private String display_value;

        private int current_meter_mode_index = -1;

        List<ObserverBase> observers = new List<ObserverBase>();

        /*********************************************************************/

        /**
         * read in given byte
         * @param: byte out of the stream
         * @return: 1 - byte accepted, 0 - failed
         */
        public int parse(byte inb)
        {
            int pos = (inb & 0xf0) >> 4;
            
            if (state + 1 != pos || pos < 1 || pos > 14)
            {
                state = 0;
                return 0;
            }

            state++;

            data[pos-1] = (byte)(inb & 0x0f);

            if (pos == 14)
            {                
                translate();
                state = 0;
            }

            return 1;
        }

        /**
         * read in given bytes
         * @param: bytes out of the stream
         * @return: 1 - bytes accepted, 0 - failed
         */
        public int parse(byte[] inb)
        {
            foreach (byte b in inb)
            {
                if (1 != parse(b))
                {
                    return 0;
                }
            }
            return 1;
        }

        /**
         * translate the complete data stream into meter attributes by
         * detecting the state/value of
         * - autorange
         * - data hold
         * - rel (relative measurement)
         * - min/max measurement
         * - meter mode
         * - current type (if applicable)
         * - unit of measured value
         * - measured value
         */
        private void translate()
        {
            bool dez_point = false, minus = false;
            valid_data = false;

            //System.Console.WriteLine("Value 12:" + data[12]);

            testAndSet(ref data[0], BitValues.BIT1VAL, ref auto_range);
            testAndSet(ref data[11], BitValues.BIT0VAL, ref data_hold);
            testAndSet(ref data[11], BitValues.BIT1VAL, ref relative_measurement);
            testAndSet(ref data[13], BitValues.BIT0VAL, ref max_value);
            testAndSet(ref data[13], BitValues.BIT3VAL, ref min_value);

            testAndSet(ref data[1], BitValues.BIT3VAL, ref minus);

            display_value = minus ? "-" : "";

            for (int i = 1; i < 8; i+=2)
            {
                testAndSet(ref data[i], BitValues.BIT3VAL, ref dez_point);

                if (dez_point)
                    display_value += ".";
                display_value += getDisplayValue(data[i], data[i + 1]);
            }

            if (detectMeterMode())
            {
                valid_data = true;
                foreach (ObserverBase observer in observers)
                {
                    observer.Update();
                }
            }            
        }
        
        /**
         * detect the current meter mode depending of the data
         * @return true if meter mode could be detected, else false
         */
        private bool detectMeterMode()
        {

            if (current_meter_mode_index != -1)
            {
                if (meter_mode_descriptions[current_meter_mode_index].test(ref data))
                {
                    return true;
                }
            }
            current_meter_mode_index = -1;

            for (int i = 0; i < meter_mode_descriptions.Count; i++)
            {                
                if (meter_mode_descriptions[i].test(ref data)) {

                    current_meter_mode_index = i;
                    return true;
                }
            }

            return false;
        }

        /**
         * convert the given nibbles to their textual representation         
         * @param first byte of the pair
         * @param second byte of the pair
         * @return textual representation of the shown display element
         */
        private String getDisplayValue(byte high_b, byte low_b)
        {
            high_b &= 0x0f;
            low_b &= 0x0f;

            switch ((DisplayValues)(high_b << 4) + low_b)
            {
                case DisplayValues.D_ZERO:
                    return "0";
                case DisplayValues.D_ONE:
                    return "1";
                case DisplayValues.D_TWO:
                    return "2";
                case DisplayValues.D_THREE:
                    return "3";
                case DisplayValues.D_FOUR:
                    return "4";
                case DisplayValues.D_FIVE:
                    return "5";
                case DisplayValues.D_SIX:
                    return "6";
                case DisplayValues.D_SEVEN:
                    return "7";
                case DisplayValues.D_EIGHT:
                    return "8";
                case DisplayValues.D_NINE:
                    return "9";
                case DisplayValues.D_EMPTY:
                    return "";
                case DisplayValues.D_L:
                    return "L";
                case DisplayValues.D_MINUS:
                    return "-";
                default:
                    return "E";
            }                        
        }

        /**
         * perform check and set operation for given parameters
         * it tests the given byte for a logical 1 on given bit-position
         * if there is a 1 the bit is cleared and the given output parameter
         * is set
         * @param [in, out] data - the byte to check and change if needed
         * @param [in] bit - the position of the byte to check for 1
         * @param [in, out] attribut - the place to store the result
         */
        private void testAndSet(ref byte data, BitValues bit_val, ref bool attribut)
        {

            if ((data & (byte)bit_val) != 0)
            {
                data -= (byte)bit_val;
                attribut = true;
            }
            else
            {
                attribut = false;
            }
        }
        /*******************************************************************/
        public GVA18BProtocolDecoder() {}

        /*******************************************************************/

        /**
        * add an external observer
        * @param observer object to be added
        */
        public void Attach(ObserverBase observer)
        {
            observers.Add(observer);
        }

        /**
        * remove an external observer
        * @param observer object to be remove
        * @return true on success (observer removed), else false
        */
        public bool Detach(ObserverBase observer)
        {
            return observers.Remove(observer);
        }


        /*******************************************************************/

        /**
        * check if valid data available
        * @return true if valid data available, else false
        */
        public bool hasValidData()
        {
            return valid_data;
        }

        /**
         * check if autorange is enabled
         * @return true if autorange is enabled, else false
         */
        public bool isAutoRange()
        {
            return auto_range;
        }

        /**
         * check if maximum value is enabled
         * @return true if maximum value is enabled, else false
         */
        public bool isMaxValue()
        {
            return max_value;
        }

        /**
         * check if minimum value is enabled
         * @return true if minimum value is enabled, else false
         */
        public bool isMinValue()
        {
            return min_value;
        }

        /**
         * check if data hold is enabled
         * @return true if data hold is enabled, else false
         */
        public bool isHold()
        {
            return data_hold;
        }

        /**
         * check if relative measurement is enabled
         * @return true if relative measurement is enabled, else false
         */
        public bool isRel()
        {
            return relative_measurement;
        }

        /**
         * check for valid value on display
         * @return true if display content could be parsed to double, else false
         */
        public bool isNumber() {
            double dummy;
            return double.TryParse(display_value, out dummy);
        }

        /**
         * check for overflow detected by meter
         * @return true if overflow detected, else false
         */
        public bool isOverflow()
        {
            return display_value.Contains("0L");
        }

        /**
         * get the name for the selcted meter mode
         * @return textual representation of selected meter mode {voltage|current|...}
         */
        public String getMeterModeName()
        {
            if (current_meter_mode_index >= 0)
            {
                return meter_mode_descriptions[current_meter_mode_index].Name;
            }
            return "";
        }

        /**
         * get the textual representation of unit for the selected meter mode
         * @return unit as string {V,mV,nF,...}
         */
        public String getMeterModeUnit()
        {
            if (current_meter_mode_index >= 0)
            {
                return meter_mode_descriptions[current_meter_mode_index].Unit;
            }
            return "";
        }

        /**
         * get the selcted type of current if applicable
         * @return textual representation for type of current {DC|AC|""}
         */
        public String getMeterModePolarity()
        {
            if (current_meter_mode_index >= 0)
            {
                return meter_mode_descriptions[current_meter_mode_index].Polarity;
            }
            return "";
        }

        /**
         * get the value shown on the display as double
         * @return parsed double-value
         */
        public double Value()
        {
            return double.Parse(display_value);
        }

        /**
         * get the value shown on the display as string
         * @return textual representation of the display content
         */
        public String ValueStr()
        {
            return display_value;
        }

    }
}
